//*****************************************************************************
//
// main.c - Main file for nRF8001 on TIVA Microcontroller
//
// Copyright (c) 2012-2013 Texas Instruments Incorporated.  All rights reserved.
// Software License Agreement
// 
// Texas Instruments (TI) is supplying this software for use solely and
// exclusively on TI's microcontroller products. The software is owned by
// TI and/or its suppliers, and is protected under applicable copyright
// laws. You may not combine this software with "viral" open-source
// software in order to form a larger program.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND WITH ALL FAULTS.
// NO WARRANTIES, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT
// NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. TI SHALL NOT, UNDER ANY
// CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
// DAMAGES, FOR ANY REASON WHATSOEVER.
// 
// This is part of revision 1.0 of the EK-TM4C123GXL Firmware Package.
//
//*****************************************************************************

#include <stdint.h>
#include <stdbool.h>
#include <math.h>
#include <time.h>
#include "inc/hw_types.h"
#include "inc/hw_memmap.h"
#include "inc/hw_hibernate.h"
#include "driverlib/fpu.h"
#include "driverlib/gpio.h"
#include "driverlib/hibernate.h"
#include "driverlib/interrupt.h"
#include "driverlib/pin_map.h"
#include "driverlib/rom.h"
#include "driverlib/sysctl.h"
#include "driverlib/systick.h"
#include "driverlib/uart.h"
#include "drivers/rgb.h"
#include "drivers/buttons.h"
#include "main.h"
#include "hw_ints.h"
#include "hal_aci_tl.h"
#include "spi.h"

#include "hal_aci_tl.h"
#include "aci.h"
#include "aci_evts.h"
#include "aci_cmds.h"
#include "lib_aci.h"

extern uint8_t rdynFlag;

//aci_struct that will contain
// total initial credits
// current credit
// current state of the aci (setup/standby/active/sleep)
// open remote pipe pending
// close remote pipe pending
// Current pipe available bitmap
// Current pipe closed bitmap
// Current connection interval, slave latency and link supervision timeout
// Current State of the the GATT client (Service Discovery)
// Status of the bond (R) Peer address
static struct aci_state_t aci_state;
static hal_aci_evt_t aci_data;
static hal_aci_data_t aci_cmd;

//*****************************************************************************
//
//! \addtogroup example_list
//! <h1>EK-TM4C123GXL Quickstart Application (qs-rgb)</h1>
//!
//! A demonstration of the Tiva C Series LaunchPad (EK-TM4C123GXL)
//! capabilities.
//!
//! Press and/or hold the left button to traverse towards the red end of the
//! ROYGBIV color spectrum.  Press and/or hold the right button to traverse
//! toward the violet end of the ROYGBIV color spectrum.
//!
//! If no input is received for 5 seconds, the application will start
//! automatically changing the color displayed.
//!
//! Press and hold both left and right buttons for 3 seconds to enter
//! hibernation.  During hibernation, the last color displayed will blink
//! for 0.5 seconds every 3 seconds.
//!
//! The system can also be controlled via a command line provided via the UART.
//! Configure your host terminal emulator for 115200, 8-N-1 to access this
//! feature.
//!
//! - Command 'help' generates a list of commands and helpful information.
//! - Command 'hib' will place the device into hibernation mode.
//! - Command 'rand' will initiate the pseudo-random color sequence.
//! - Command 'intensity' followed by a number between 0 and 100 will set the
//! brightness of the LED as a percentage of maximum brightness.
//! - Command 'rgb' followed by a six character hex value will set the color.
//! For example 'rgb FF0000' will produce a red color.
//
//*****************************************************************************

//*****************************************************************************
//
// Entry counter to track how int32_t to stay in certain staging states before
// making transition into hibernate.
//
//*****************************************************************************
static volatile uint32_t ui32HibModeEntryCount;

//*****************************************************************************
//
// Application state structure.  Gets stored to hibernate memory for
// preservation across hibernate events.
//
//*****************************************************************************
volatile tAppState g_sAppState;

//*****************************************************************************
//
// The error routine that is called if the driver library encounters an error.
//
//*****************************************************************************
#ifdef DEBUG
void
__error__(char *pcFilename, uint32_t ui32Line)
{
}
#endif


//*****************************************************************************
//
// Uses the fColorWheelPos variable to update the color mix shown on the RGB
//
// ui32ForceUpdate when set forces a color update even if a color change
// has not been detected.  Used primarily at startup to init the color after
// a hibernate.
//
// This function is called by the SysTickIntHandler to update the colors on
// the RGB LED whenever a button or timeout event has changed the color wheel
// position.  Color is determined by a series of sine functions and conditions
//
//*****************************************************************************
void
AppRainbow(uint32_t ui32ForceUpdate)
{
    static float fPrevPos;
    float fCurPos;
    float fTemp;
    volatile uint32_t * pui32Colors;

    pui32Colors = g_sAppState.ui32Colors;
    fCurPos = g_sAppState.fColorWheelPos;


    if((fCurPos != fPrevPos) || ui32ForceUpdate)
    {
        //
        // Preserve the new color wheel position
        //
        fPrevPos = fCurPos;

        //
        // Adjust the BLUE value based on the control state
        //
        fTemp = 65535.0f * sinf(fCurPos);
        if(fTemp < 0)
        {
            pui32Colors[GREEN] = 0;
        }
        else
        {
            pui32Colors[GREEN] = (uint32_t) fTemp;
        }


        //
        // Adjust the RED value based on the control state
        //
        fTemp = 65535.0f * sinf(fCurPos - APP_PI / 2.0f);
        if(fTemp < 0)
        {
            pui32Colors[BLUE] = 0;
        }
        else
        {
            pui32Colors[BLUE] = (uint32_t) fTemp;
        }


        //
        // Adjust the GREEN value based on the control state
        //
        if(fCurPos < APP_PI)
        {
            fTemp = 65535.0f * sinf(fCurPos + APP_PI * 0.5f);
        }
        else
        {
            fTemp = 65535.0f * sinf(fCurPos + APP_PI);
        }
        if(fTemp < 0)
        {
            pui32Colors[RED] = 0;
        }
        else
        {
            pui32Colors[RED] = (uint32_t) fTemp;
        }

        //
        // Update the actual LED state
        //
        RGBColorSet(pui32Colors);
    }

}

//*****************************************************************************
//
// Called by the NVIC as a result of SysTick Timer rollover interrupt flag
//
// Checks buttons and calls AppButtonHandler to manage button events.
// Tracks time and auto mode color stepping.  Calls AppRainbow to implement
// RGB color changes.
//
//*****************************************************************************
void
SysTickIntHandler(void)
{
    //
    //    Set the RGB Color based on current control variable value.
    //
    AppRainbow(0);
}

//*****************************************************************************
//
// Configure the UART and its pins.  This must be called before UARTprintf().
//
//*****************************************************************************
void
ConfigureUART(void)
{
    //
    // Enable the GPIO Peripheral used by the UART.
    //
    ROM_SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOA);

    //
    // Enable UART0
    //
    ROM_SysCtlPeripheralEnable(SYSCTL_PERIPH_UART0);

    //
    // Configure GPIO Pins for UART mode.
    //
    ROM_GPIOPinConfigure(GPIO_PA0_U0RX);
    ROM_GPIOPinConfigure(GPIO_PA1_U0TX);
    ROM_GPIOPinTypeUART(GPIO_PORTA_BASE, GPIO_PIN_0 | GPIO_PIN_1);

    //
    // Use the internal 16MHz oscillator as the UART clock source.
    //
    UARTClockSourceSet(UART0_BASE, UART_CLOCK_PIOSC);

    //
    // Initialize the UART for console I/O.
    //
    UARTStdioConfig(0, 115200, 16000000);
}

//*****************************************************************************
//
// Main function performs init and manages system.
//
// Called automatically after the system and compiler pre-init sequences.
// Performs system init calls, restores state from hibernate if needed and
// then manages the application context duties of the system.
//
//*****************************************************************************
int
main(void)
{
    //
    // Enable stacking for interrupt handlers.  This allows floating-point
    // instructions to be used within interrupt handlers, but at the expense of
    // extra stack usage.
    //
    ROM_FPUEnable();
    ROM_FPUStackingEnable();

    //
    // Set the system clock to run at 40Mhz off PLL with external crystal as
    // reference.
    //
    ROM_SysCtlClockSet(SYSCTL_SYSDIV_5 | SYSCTL_USE_PLL | SYSCTL_XTAL_16MHZ |
                       SYSCTL_OSC_MAIN);

    //
	// Enable and Initialize the UART.
	//
	ConfigureUART();
	UARTprintf("Welcome to the Tiva C Series TM4C123G LaunchPad with nRF8001!\n");

    //
    // Initialize the RGB LED. AppRainbow typically only called from interrupt
    // context. Safe to call here to force initial color update because
    // interrupts are not yet enabled.
    //
    RGBInit(0);
    RGBIntensitySet(g_sAppState.fIntensity);

    // Display Red Color
	g_sAppState.ui32Colors[0] = 0;
	g_sAppState.ui32Colors[1] = 0;
	g_sAppState.ui32Colors[2] = 0xFFFF;
    AppRainbow(1);
    RGBEnable();

    //This will reset the nRF8001. ACI Device Started Event is generated by the nRF8001 device
	//as soon as the reset is complete
	hal_aci_tl_init();

    //
    // Initialize the SysTick interrupt to process colors and buttons.
    //
    SysTickPeriodSet(SysCtlClockGet() / APP_SYSTICKS_PER_SEC);
    SysTickEnable();
    SysTickIntEnable();
    IntMasterEnable();

    reset_nrf8001();

    //
    // spin forever and wait for carriage returns or state changes.
    //
    while(1)
    {
    	// We enter the if statement only when there is a ACI event available to be processed
		if (lib_aci_event_get(&aci_state, &aci_data))
		{
			aci_evt_t * aci_evt;

			aci_evt = &aci_data.evt;
			switch(aci_evt->evt_opcode)
			{
				/**
				As soon as you reset the nRF8001 you will get an ACI Device Started Event
				*/
				case ACI_EVT_DEVICE_STARTED:
				{

				  aci_state.data_credit_available = aci_evt->params.device_started.credit_available;
				  switch(aci_evt->params.device_started.device_mode)
				  {
					case ACI_DEVICE_SETUP:

					  /**
					  Device is in the setup mode
					  */
					  UARTprintf("Evt Device Started: Setup\n");
					  aci_cmd.buffer[0] = 2;    //Length of ACI command
					  aci_cmd.buffer[1] = 0x01; //Command - Test
					  aci_cmd.buffer[2] = ACI_TEST_MODE_DTM_UART; //Command parameter
					  hal_aci_tl_send(&aci_cmd);
					  break;

					case ACI_DEVICE_TEST:
						UARTprintf("Evt Device Started: Test\n");

					  break;
				  }
				}
				break; //ACI Device Started Event
			  case ACI_EVT_CMD_RSP:
				//If an ACI command response event comes with an error -> stop
				if (ACI_STATUS_SUCCESS != aci_evt->params.cmd_rsp.cmd_status)
				{
				  //ACI ReadDynamicData and ACI WriteDynamicData will have status codes of
				  //TRANSACTION_CONTINUE and TRANSACTION_COMPLETE
				  //all other ACI commands will have status code of ACI_STATUS_SCUCCESS for a successful command
				  //Serial.print(F("ACI Command "));
				  //Serial.println(aci_evt->params.cmd_rsp.cmd_opcode, HEX);
				  //Serial.println(F("Evt Cmd respone: Error. In an while(1); loop"));
				  while (1);
				}
				break;
			}
		}
		else
		{
		//Serial.println(F("No ACI Events available"));
		// No event in the ACI Event queue.
		// Arduino can go to sleep
		// Wakeup from sleep from the RDYN line
		}
    }
}
